/*
 * ocache_mem.c --- memory handling procedures for object caching at KOTETU.
 *      by k-chinen@is.aist-nara.ac.jp, 2000, 2001.
 *
 *  this routines is splitted from xsk.c of kotetu-1.2.
 *
 * $Id$
 */
/*
 * xsk.c --- sk.c + skcache.c
 *      by k-chinen@is.aist-nara.ac.jp, 2000, 2001.
 *
 * $Id$
 */
/*
 * sk.c --- skip list procedures
 *              for KOTETU
 *      by k-chinen@is.aist-nara.ac.jp, 2000
 */
/*
 * skcache.c --- object cache procedures using skip list and its bucket
 *              for KOTETU
 *      by k-chinen@is.aist-nara.ac.jp, 2000
 */
/*

 terminology:
  entry   - object entry. store properties like PCB of TCP/IP
            or super block of filesystem.
  frag    - fragment of object cache. object cache is devided little block
            and stored its list.
            you can access entity of object by drawing its list.
  bucket  - list of entry.  it was controled as skip list.
 
*/
/*
 NOTE:

 hash:
    this program use meny hash to assign object to structures in several level
    and identify among objects.

 on the memory:
    cache consists set of buckets and frag lists in entry.

        bucket[0]
            entry[0]    -> frag[0], frag[1], frag[2] ...
            entry[1]
            entry[2]
            ...

        bucket[1]
            entry[0]
            entry[1]
            entry[2]
            ...

 on the disk:
    cache consists dirctories according to hash.

        aa/bb/cc-dd
        
        aa - first level dir. with first hash (so it is bucket)
        bb - second level dir. with second hash
        cc-dd - file with third and fourth hash.

 */

/*
 * See also:
 *   William Pugh, "skiplist cookbook",
 *   UMIACS-TR-89-72.1, July 1989
 */

/*
 * The algorithm on the original text was described by Pascal like
 * langauage.  Subscription were 1..maxlevel or 1..level .
 * Since this is C program,  This file describes subscription as
 * 0..maxlevel-1 and 0..level-1,
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <limits.h>
#include <pthread.h>
#include <time.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/uio.h>
#include <dirent.h>

#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif

#include <sys/mman.h>
#include <assert.h>
#include <errno.h>

#include "range.h"
#include "msg.h"
#include "trace.h"
#include "misc.h"
#include "mem.h"
#include "ocache.h"
#include "sim_stat.h"
#include "sim_hist.h"
#include "mlist.h"

#ifndef FRAG_NTRY
#define FRAG_NTRY   (32)
#endif


extern off_t require_total_memory_size;

/*
 * arround 256, follwing numbers are known as prime numbers.
 *  241 251 257 263 269 271
 *
 */

int radix_h1 =   241;
int radix_h2 =   239;
int radix_h3 =   233;
int radix_h4 =   251;

skcache_entry_t *NIL;
int nbucket;
bucket_t *skcache_bucket_base;

int nentry_bucket=-1;

extern pthread_mutex_t stat_lock;


volatile int    compaction_count   = 0;
sim_stat_rec    compaction_time;
static long     compaction_time_count[K(2)];
sim_stat_rec    checkmem_time;
static long     checkmem_time_count[K(2)];

volatile int    msweeper_count     = 0;
sim_stat_rec    msweep_time;
sim_stat_rec    msweep_time_short;
static long     msweep_time_count[K(10)];
static long     msweep_time_short_count[K(10)];


int require_radix_h2;
int require_radix_h3;
off_t require_dir_size;
int require_disk_sweep_interval;


extern pthread_mutex_t config_lock;

int require_keeptime_token_val  = 1;
int skcache_keeptime_token_init = 1;
int skcache_keeptime_token      = 1;
int skcache_keeptime_ave_range  = 3;
int skcache_mem_keep_min    = 12;
int skcache_mem_keep_max    = 3600*4;   /* 4 hours, 1/6 day */
int skcache_mem_keep_cur    = 3600;
int skcache_entry_keep_min  = 120;
int skcache_entry_keep_cur  = 86400;    /* 1 day */
int skcache_entry_keep_max  = 86400*7;  /* 1 week */

int skcache_signature_refresh = DSIG_REFRESH;



extern volatile long long swap_count;
extern volatile long long swap_size;

int require_swap_token_val;
int require_swap_token_interval;

#if 0
extern volatile int swapout_token_init;
extern volatile int swapout_token;
extern volatile int swapout_token_interval;
#endif
extern volatile int disk_sweep_interval;

time_t min_date;

pthread_mutex_t dir_mutex = PTHREAD_MUTEX_INITIALIZER;
int dir_num=0;
int *dir_lock;
u_int32_t *dir_sig;
off_t *dir_sz;
int *dir_count;
int *dir_time;
int *dir_lack;
int dir_conflict = 0;


pthread_mutex_t frag_lock = PTHREAD_MUTEX_INITIALIZER;
int z_frag_freeliss[5] = {0,1,2,3,4};
skcache_frag_t *z_frag_freelist = NULL;
int z_frag_freelisu[5] = {0,1,2,3,4};
int frag_dummies[L_BUF] = {0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15, };
int frag_all = -1, frag_use = -1, frag_free = -1;
int frag_hhwater, frag_hwater, frag_lwater;
int frag_size = -1;

int require_entry_hhwater;
int require_entry_hwater;
int require_entry_lwater;
int require_frag_hhwater;
int require_frag_hwater;
int require_frag_lwater;


sim_hist_rec fratio_hist;
static long fratio_hist_area[128];



TRACE_REGIST_NIC("objcachemem");


#if 1
/* nanosleep is useful, because it is MT-safe and supports very short time.
 * however, nsec is too short for network I/O.
 * then, msleep is made.
 */
/*
inline
*/
static
int
msleep(int x)
{
    struct timespec w;
    long long tw;

    /* convert msec to nsec */
    tw = x * 1000000;
    w.tv_nsec = tw % 1000000000;
    w.tv_sec =  tw / 1000000000;

    nanosleep(&w, NULL);
    
    return 0;
}
#endif



void
sk_rawdump(FILE* fp, skiplist_t *list, char* (fn)(void*))
{
    skiplist_cell_t *cur;
    int c, i;

#if 0
    for(i=0;i<list->level;i++) {
        fprintf(stdout, " h %4d: %p\n", i, list->header->skip[i]);
    }
#endif

    fprintf(fp, "--- list addr %p, num %d, func %p\n",
        list, list->num, fn);

    c = 0;
    cur = list->header->skip[0];
    if(fn)
        while(cur != list->term) {
            fprintf(fp, " %4d: %p : %p %s // %p\n",
                c, cur, cur->entity, fn(cur->entity), cur->skip[0]);
            cur = cur->skip[0];
            c++;
        }
    else
        while(cur != list->term) {
            fprintf(fp, " %4d: %p : %p // %p\n",
                c, cur, cur->entity, cur->skip[0]);
            cur = cur->skip[0];
            c++;
        }
}


void
sk_dump(skiplist_t* list, FILE*fp, char* (fn)(void*) )
{
    skiplist_cell_t* cur;
    int i;
    char label[256];
    char *p;
    int k;
    char tmp[L_BUF];
    char line[L_BUF];

    if(list==NULL) {
        return;
    }

    fprintf(fp, "skiplist %p, num %d, level %d/%d\n",
        list, list->num, list->level, list->maxlevel);

    line[0] = '\0';
    sprintf(tmp, "%-38s %-10s %-10s ", "header", "element", "entity");
    strcat(line, tmp);
    for(i=0;i<=list->level;i++) {
        if(list->header->skip[i]==NULL) {
            sprintf(tmp, "%-10s ", "NIL");
        }
        else
        if(list->header->skip[i]==list->term) {
            sprintf(tmp, "%-10s ", "TERM");
        }
        else {
            sprintf(tmp, "0x%08x ", list->header->skip[i]);
        }
        strcat(line, tmp);
    }
    /*
    fputs(line, fp);
    */
    fprintf(fp, "%s\n", line);

    cur = list->header->skip[0];
    while(cur && cur!=list->term) {
        line[0] = '\0';
        if(fn!=NULL) {
            p = fn(cur->entity);
            if(p) {
                strcpy(label, p);
            }
            else {
                strcpy(label, "*strange*");
            }
        }
        else {
            strcpy(label, "*NONE*");
        }

        sprintf(tmp, "%-38s 0x%08x 0x%08x ", label, cur, cur->entity);
        strcat(line, tmp);

        for(i=0;i<=list->level;i++) {
            if(cur->skip[i]==NULL) {
                sprintf(tmp, "%-10s ", "NIL");
            }
            else
            if(cur->skip[i]==list->term) {
                sprintf(tmp, "%-10s ", "TERM");
            }
            else {
                sprintf(tmp, "0x%08x ", cur->skip[i]);
            }
            strcat(line, tmp);
        }
        /*
        fputs(line, fp);
        */
        fprintf(fp, "%s\n", line);

        if(cur==list->term) {
            break;
        }
        if(cur==cur->skip[0]) {
            fprintf(fp, "*** looped\n");
            break;
        }

        cur = cur->skip[0];
    }

    fprintf(fp, "%-38s 0x%08x\n", "TERM", list->term);

#if 0
    fprintf(fp, "%-24s %-10s ", "NIL", "NIL");
    fprintf(fp, "\n\n");
#endif


}

void
bk_dump(bucket_t *bk, skiplist_t* list, FILE*fp, char* (fn)(void*) )
{
    char *pcur;
    skiplist_cell_t* cur;
    int i;
    char label[256];
    char *p;
    int j, k;

    if(list==NULL) {
        return;
    }

    fprintf(fp, "### list %p, num %d, level %d/%d, term %p\n",
        list, list->num, list->level, list->maxlevel, list->term);

#ifndef DEBUG
    return;
#endif

#ifdef DEBUG
    fprintf(fp, "#  first %p last %pd\n#\n",
        bk->cell_first, bk->cell_last);

#if 0
    if(fn==NULL) {
        fprintf(fp, "#%6s %8s %8s %s %p\n",
            "", "", "", "header", list->header);
        fprintf(fp, "#%6s %8s %8s ",
            "", "", "");
    }
    else {
        fprintf(fp, "#%6s %8s %8s %-14s %s %p\n",
            "", "", "", "", "header", list->header);
        fprintf(fp, "#%6s %8s %8s %-14s ",
            "", "", "", "");
    }
    if(list->level<0) {
        fprintf(fp,"*** no level ***\n");
    }
    else {

        cur = list->header;
        for(i=0;i<=list->level;i++) {
            if(cur->skip[i]==NULL) {
                fprintf(fp, "%-8s ", "NIL");
            }
            else {
                fprintf(fp, "%p ", cur->skip[i]);
            }
        }
    fprintf(fp, "\n");

    }
#endif

#if 1
    fprintf(fp, "# cup  header %p: ", bk->cup.header);
        cur = bk->cup.header;
        for(i=0;i<=list->level;i++) {
            if(cur->skip[i]==NULL) {
                fprintf(fp, "%-8s ", "NIL");
            }
            else {
                fprintf(fp, "%p ", cur->skip[i]);
            }
        }
        fprintf(fp, "\n");
    fprintf(fp, "# cup  term   %p: ", bk->cup.term);
        cur = bk->cup.term;
        for(i=0;i<=list->level;i++) {
            if(cur->skip[i]==NULL) {
                fprintf(fp, "%-8s ", "NIL");
            }
            else {
                fprintf(fp, "%p ", cur->skip[i]);
            }
        }
        fprintf(fp, "\n");

#endif


    fprintf(fp, "#\n");

    if(fn==NULL) {
        fprintf(fp, "#%6s %8s %8s %8s",
            "num", "addr", "entity", "skips..");
    }
    else {
        fprintf(fp, "#%6s %8s %8s %-14s %8s",
            "num", "addr", "entity", "label", "skips..");
    }
    fprintf(fp, "\n");

    j = -1;
    k = 0;
    pcur = (char*)bk->cell_first;
    while(pcur<=(char*)bk->cell_last) {
        cur=(skiplist_cell_t*)pcur;

        j++;
        if(cur->entity==NULL) {
            goto next;
        }
#if 0
        fprintf(fp, "%p %p ", cur->entity, cur);
#endif
        if(fn!=NULL) {
            p = fn(cur->entity);
            if(p) {
                strcpy(label, p);
            }
            else {
                strcpy(label, "*strange*");
            }
            fprintf(fp, "= %6d %6d %p %p %-14s ",
                j, k, cur, cur->entity, label);
        }
        else {
            fprintf(fp, "= %6d %6d %p %p ",
                j, k, cur, cur->entity);
        }

        for(i=0;i<=list->level;i++) {
            if(cur->skip[i]==NULL) {
                fprintf(fp, "%-8s ", "x");
            }
            else {
                fprintf(fp, "%p ", cur->skip[i]);
            }
        }
        fprintf(fp, "\n");
        k++;
#if 0
        if(k>list->num+3) {
            break;
        }
#endif

next:
        pcur += bk->elemsize;
    }

    fprintf(fp, "\n\n");
    fflush(fp);

#endif
}




static
int
randomlevel(int max, double p)
{
    int lv;
    int w = 4097;
    double v;

    lv = 0;
#if 0
    while( v=((double)(rand()%w))/w, v < p && lv < max) {
        lv++;
    }
#endif
    while(drand48() < p && lv < max) {
        lv++;
    }

    return lv;
}




/*
 * cmp - compare
 */
skiplist_cell_t *
sk_find_listpos(skiplist_t *list, void* newface,
        int (cmp)(void*,void*))
{
    register skiplist_cell_t *x;
    register int i, diff;
    skiplist_cell_t *ret;

    ret = NULL;

    x = list->header;
    for(i=list->level; i>=0; i--) {
        while(1) {
            if(x->skip[0]==NULL || x->skip[i]==list->term) {
                break;
            }
            diff = cmp(x->skip[i]->entity, newface);
            if(diff>=0) {
                break;
            }
            x = x->skip[i];
        }
        /* bingo */
        if(diff==0) {
            ret = x->skip[i];
            goto out;
        }
    }
    x = x->skip[0];

    if(x!=list->term && cmp(x->entity, newface)==0) {
        ret = x;
    }

out:
    return ret;
}

void*
sk_find(skiplist_t *list, void* newface,
        int (cmp)(void*,void*))
{
    register skiplist_cell_t *x;
    register int i, diff;
    void *ret;

    ret = NULL;

    x = list->header;
    for(i=list->level; i>=0; i--) {
        while(1) {
            if(x->skip[0]==NULL || x->skip[i]==list->term) {
                break;
            }
            diff = cmp(x->skip[i]->entity, newface);
            if(diff>=0) {
                break;
            }
            x = x->skip[i];
        }
        /* bingo */
        if(diff==0) {
            ret = x->skip[i]->entity;
            goto out;
        }
    }
    x = x->skip[0];

    if(x!=list->term && cmp(x->entity, newface)==0) {
        ret = x->entity;
    }

out:
    return ret;
}


/*
 * if alreay assigned, not new allocation.
 * cmp - compare
 */
void*
sk_assign(skiplist_t *list, void* newface,
        skiplist_cell_t **flist,
        int (cmp)(void*,void*),
        int* st, int opt)
{
    skiplist_cell_t *update[SKL_MAXLEVEL];
    register skiplist_cell_t *x;
    register int i;
    int newlevel;
    skiplist_cell_t *nextfree;
    skiplist_cell_t *ret;
    int c;
    int diff;

#if 1
    if(newface==NULL) {
        Error("don't assign NULL to skiplist\n");

        ret = NULL;
        *st = -1;
        goto out;
    }
#endif

    if(*flist==NULL) {
        fprintf(stderr, "sk_assign: no free cell in entity list (opt %d)\n", opt);
        ret = NULL;
        *st = -1;
        goto out;
    }

    x = list->header;
    for(i=list->level; i>=0; i--) {
        c = 0;

        while(1) {
            if(x->skip[i]==NULL || x->skip[i]==list->term
                        || x->skip[i]->entity==NULL) {
                diff = INT_MAX;
                break;
            }
            diff = cmp(x->skip[i]->entity, newface);

            if(diff>=0) {
                break;
            }
            x = x->skip[i];
            c++;
        }
        update[i] = x;
    }
    x = x->skip[0];

assert(x && x->entity);

    if(cmp(x->entity, newface)==0) {
        ret = x->entity;
        *st = 1;
    }
    else {
        newlevel = randomlevel(list->maxlevel, 0.5);

assert(newlevel<=list->maxlevel);

        if(newlevel > list->level) {
            for(i=list->level+1;i<=newlevel;i++) {
                update[i] = list->header;
            }
            list->level = newlevel;
        }

        if(*flist==NULL) {
            fprintf(stderr, "no free cell (opt %d)\n", opt);
            ret = NULL;
            *st = -1;
            goto out;
        }

        nextfree = (*flist)->skip[0];

        x = *flist;
        *flist = nextfree;

        x->entity = newface;
        list->num++;

        ret = x->entity;
        *st = 0;

        for(i=0;i<=newlevel;i++) {
            x->skip[i] = update[i]->skip[i];
            update[i]->skip[i] = x;
        }

    }

out:
    return ret;
}

int
sk_insert(skiplist_t *list, void* newface,
        skiplist_cell_t **flist,
        int (cmp)(void*,void*), int (ovw)(void**,void*), int opt)
{
    skiplist_cell_t *update[SKL_MAXLEVEL];
    register skiplist_cell_t *x;
    register int i;
    int newlevel;
    skiplist_cell_t *nextfree;
    int c;
    int diff;

    x = list->header;
    for(i=list->level; i>=0; i--) {
        c = 0;

        while(1) {
            if(x->skip[0]==NULL || x->skip[0]==list->term) {
                break;
            }
            diff = cmp(x->skip[i]->entity, newface);
            if(diff>=0) {
                break;
            }
            x = x->skip[i];
            c++;
        }
        update[i] = x;
    }
    x = x->skip[0];

assert(x && x->entity);

    if(cmp(x->entity, newface)==0 && ovw) {
        ovw(&x->entity, newface);
    }
    else {
        newlevel = randomlevel(list->maxlevel, 0.5);

assert(newlevel<=list->maxlevel);

        if(newlevel > list->level) {
            for(i=list->level+1;i<=newlevel;i++) {
                update[i] = list->header;
            }
            list->level = newlevel;
        }

        if(*flist==NULL) {
            fprintf(stderr, "no free cell (opt %d)\n", opt);
            return -1;
        }

        nextfree = (*flist)->skip[0];

        x = *flist;
        *flist = nextfree;

        x->entity = newface;
        list->num++;

        for(i=0;i<=newlevel;i++) {
            x->skip[i] = update[i]->skip[i];
            update[i]->skip[i] = x;
        }

    }

    return 0;
}

/*
 * arguments:
 *      cmp - compare
 *      rem - remove
 * return:
 *      NULL    fail    
 *      !NULL   success
 */
void*
sk_delete(skiplist_t *list, void* newface,
        skiplist_cell_t **flist,
        int (cmp)(void*,void*), int (rem)(void*), int opt)
{
    skiplist_cell_t *update[SKL_MAXLEVEL];
    skiplist_cell_t *x;
    int i;
    void *ret;
    int diff;
    int chk;
    int z;
    char* skcache_label4_ref(void *pp);

    ret = NULL;

    x = list->header;
    z = 0;
    for(i=list->level; i>=0; i--) {
        while(1) {
            if(x->skip[i]==NULL || x->skip[i]==list->term) {
                break;
            }
            if(x->skip[i]->entity==NULL) {
                Error("sk_delete: found empty entity");
#if 1
    sk_dump(list, stderr, skcache_label4_ref);
#endif
                break;
            }
            diff = cmp(x->skip[i]->entity, newface);
            if(diff>=0) {
                break;
            }
            x = x->skip[i];
        }
        update[i] = x;
    }
    x = x->skip[0];

    if(x==NULL) {
Error("sk_delete: meat NULL element");
        return ret;
    }
    if(x==list->term) {
Error("sk_delete: reach to end");
        return ret;
    }
    if(x->entity==NULL) {
Error("sk_delete: meat NULL entity");
        return ret;
    }

    if(x->entity == (skcache_entry_t*)newface
      || cmp(x->entity, newface)==0) {
        if(rem) {
            chk = rem(x->entity);
            if(chk) {
Error("sk_delete: rem() return !0 (%d)", chk);
                return ret;
            }
        }

        for(i=0;i<=list->level;i++) {
            if(update[i]->skip[i]==x) {
                update[i]->skip[i] = x->skip[i];
            }
        }

        while(list->level>0 &&
                list->header->skip[list->level]==list->term) {
            list->level--;
        }

        if(x->entity==NULL) {
Error("sk_delete: entity is NULL");
        }

        ret = x->entity;
        x->entity = NULL;
        for(i=1;i<=list->maxlevel;i++) {
            /*
            x->skip[i] = NULL;
            */
            x->skip[i] = list->term;
        }
        x->skip[0] = *flist;
        *flist = x;

        list->num--;
    }
#if 1
    else {
        Trace("sk_delete: not found entry %p\n", newface);
    }
#endif

#if 0
    if(ret==NULL) {
Error("sk_delete: return is NULL");
    }
#endif

    return ret;
}

/*
 * ack - decide delete or not
 *          that function have to return not zero for delete
 */
int
sk_mapdelete(skiplist_t *list,
        skiplist_cell_t **flist,
        int (ack)(void*, void*), void *opt)
{
    skiplist_cell_t *update[SKL_MAXLEVEL];
    skiplist_cell_t *x, *y, *n;
    int i;
    int ret;
    int diff;
    int chk;

    ret = 0;

    x = list->header;
    for(i=0;i<=list->level;i++) {
            update[i] = x;
    }

    x = x->skip[0];

    while(x!=list->term) {
        n = x->skip[0]; /* save next element */

        if(ack==NULL)
            goto delete;

        chk = ack(x->entity, opt);

        if(chk) {
delete:
            /*
             * skip current element
             * so, it is mean deletation of element
             * this routine don't care actual enitity procedure
             * (e.g, free(), unlink() or something )
             */
            for(i=0;i<=list->level;i++) {
                if(update[i]->skip[i]==x) {
                    update[i]->skip[i] = x->skip[i];
                }
            }

            /* add current element to free-list for recycle */
            y = x;
            y->entity = NULL;
            for(i=1;i<=list->maxlevel;i++) {
#if 0
                y->skip[i] = NULL;
#endif
                y->skip[i] = list->term;
            }
            y->skip[0] = *flist;
            *flist = y;

            list->num--;

            ret++;
        }
no_delete:

        for(i=0;i<=list->level;i++) {
            if(update[i]->skip[i]==x)
                update[i] = x;
        }

        x = n;
    }

    return ret;
}

skiplist_t*
sk_new(int nlevel, void* user_term)
{
    skiplist_t *this;
    skiplist_cell_t *n;
    int i, j;

    this = (skiplist_t*)MALLOC(sizeof(skiplist_t)*1);
    if(this==NULL) {
        return NULL;
    }

    if(nlevel>SKL_MAXLEVEL-1) {
        nlevel = SKL_MAXLEVEL-1;
    }

    this->maxlevel  = nlevel;
    this->level     = 0;

    n = this->header;
        n->entity = user_term;
        n->skip = (skiplist_cell_t**)MALLOC(
                    sizeof(void*)+sizeof(skiplist_cell_t**)
                    +sizeof(skiplist_cell_t*)*(nlevel+1));
        for(j=0;j<=nlevel;j++) {
            n->skip[j] = this->term;
        }

    n = this->term;
        n->entity = user_term;
        n->skip = (skiplist_cell_t**)MALLOC(
                    sizeof(void*)+sizeof(skiplist_cell_t**)
                    +sizeof(skiplist_cell_t*)*(nlevel+1));
        for(j=0;j<=nlevel;j++) {
            n->skip[j] = this->term;
        }

    return this;
}


int
bucket_show(bucket_t *bk)
{
    skiplist_cell_t *n;
    skiplist_cell_t **p;
    int i, j;

    Trace("=== bucket %p, bk# %d,\n", bk, bk->bid);

    Trace("cup  header,skip: %p\n", bk->cup.header->skip);
    n  = bk->cup.header->skip[0];
    i  = 0;
    while(n!=bk->cup.term) {
        Trace("%4d %4d %p :- %p %p\n",
            bk->bid, i,
            n, n->entity, n->skip[0]);
        n = n->skip[0];
        i++;
    }
    Trace("  have %d item(s)\n", i);

#if 0
    Trace("free list: %p\n", bk->elem_freelist);

    n  = bk->elem_freelist;
    i  = 0;
    while(n && n!=bk->cup.term) {
        Trace( "%4d %4d %p :: %p %p\n",
            bk->bid, i,
            n, n->entity, n->skip[0]);
        n = n->skip[0];
        i++;
    }
    Trace("  have %d item(s)\n", i);
#endif

    return 0;
}

int
old_bucket_show(bucket_t *bk)
{
    skiplist_cell_t *n;
    skiplist_cell_t **p;
    int i, j;

    fprintf(stdout, "=== bk %p\n", bk);

    fprintf(stdout, "cup  header,skip: %p\n", bk->cup.header->skip);
    n  = bk->cup.header->skip[0];
    i  = 0;
    while(n) {
        fprintf(stdout, "%4d: <%p> %p %p\n",
            i, n, n->entity, n->skip[0]);
        n = n->skip[0];
        i++;
    }
    fprintf(stdout,"  have %d item(s)\n", i);

    fprintf(stdout, "free list: %p\n", bk->elem_freelist);

    n  = bk->elem_freelist;
    i  = 0;
    while(n) {
        fprintf(stdout, "%4d: <%p> %p %p\n",
            i, n, n->entity, n->skip[0]);
        n = n->skip[0];
        i++;
    }
    fprintf(stdout,"  have %d item(s)\n", i);

    return 0;
}

int
bk_new(bucket_t *bk, int nlevel, int nelement, void* user_term)
{
    skiplist_cell_t *n;
    skiplist_cell_t **p;
    int i, j;
    char *pool;


    if(bk==NULL) {
        return 0;
    }

    if(nlevel>SKL_MAXLEVEL-1) {
        nlevel = SKL_MAXLEVEL-1;
    }

    bk->elemnum = nelement;
    bk->elemsize = sizeof(void*)+sizeof(skiplist_cell_t**)
            +sizeof(skiplist_cell_t*)*(nlevel+1);
    bk->totalsize = bk->elemsize*(nelement+4);

    pool = (char*)MALLOC(bk->totalsize);
    if(pool==NULL) {
        fprintf(stderr, "not enough memory !!!\n");
        return -1;
    }

#if 0
    fprintf(stderr, "nlevel %d, nelement %d\n",
            nlevel, nelement);
    fprintf(stderr, "element size %d, num %d -> total %d\n",
            bk->elemsize, nelement, bk->totalsize);
    fprintf(stderr, "pool  [%p..%p)\n", pool, pool+bk->totalsize);
#endif
#if 0
    for(i=0;i<nelement+2;i++) {
        fprintf(stdout, "%4d: <%p>\n", i, pool+bk->elemsize*i);
    }
    fflush(stdout);
#endif
    if(pthread_mutex_init(&bk->mutex, NULL)<0) {
        Error("fail pthread_mutex_init() (%d)", errno);
    }

    bk->cup.num         = 0;
    bk->cup.maxlevel    = nlevel;
    bk->cup.level       = 0;
    bk->cup.header      = (skiplist_cell_t*)pool;
    bk->cup.term        = (skiplist_cell_t*)(pool + 1*bk->elemsize);
    n = bk->cup.header;
        n->entity = user_term;
        n->skip = (skiplist_cell_t**)((char*)n+sizeof(void*)*2);
        for(j=0;j<=nlevel;j++) {
            n->skip[j] = bk->cup.term;
        }
    n = bk->cup.term;
        n->entity = user_term;
        n->skip = (skiplist_cell_t**)((char*)n+sizeof(void*)*2);
        for(j=0;j<=nlevel;j++) {
            n->skip[j] = bk->cup.term;
        }


#if 0
    fprintf(stderr, "cup  header %p\n", bk->cup.header);
    fprintf(stderr, "cup  term   %p\n", bk->cup.term);
#endif

    bk->elem_freelist = (skiplist_cell_t*)(pool+4*bk->elemsize);
#if 0
    fprintf(stderr, "elem_freelist    %p\n", bk->elem_freelist);
#endif


#if 0
#ifndef NDEBUG
    bk->cell_first = (skiplist_cell_t*)(pool+0*bk->elemsize);
    bk->cell_last  = (skiplist_cell_t*)(pool+nelement*bk->elemsize);

    fprintf(stderr, "  first..last %p..%p\n", bk->cell_first, bk->cell_last);
#endif
#endif

    n = bk->elem_freelist;
    for(i=0;i<nelement;i++) {
        if(n==NULL) {
            break;
        }
#if 0
        fprintf(stdout, "* %d, %p\n", i, n);
        fflush(stdout);
#endif
        n->entity = NULL;
        n->skip = (skiplist_cell_t**)((char*)n+sizeof(void*)*2);
        for(j=1;j<=nlevel;j++) {
/*
            n->skip[j] = NULL;
*/
            n->skip[j] = bk->cup.term;
        }
        if(i<nelement-1) {
            n->skip[0] = (skiplist_cell_t*)((char*)n+bk->elemsize);
            n = (skiplist_cell_t*)((char*)n+bk->elemsize);
        }
        else {
            n->skip[0]  = NULL;
        }
    }

#if 0
    fprintf(stdout, "cup  header,skip: %p\n", bk->cup.header->skip);
    fprintf(stdout, "free: %p\n", bk->elem_freelist);

    n  = bk->elem_freelist;
    for(i=0;i<nelement;i++) {
        if(n==NULL)
            break;

#if 0
        fprintf(stdout, "%4d: <%p> ", i, n);
        for(j=0;j<=nlevel;j++) {
            fprintf(stdout, "%d:<%p> %p ",
                j, &(n->skip[j]), n->skip[j]);
        }
        fprintf(stdout, "\n");
#endif
#if 1
        fprintf(stdout, "%4d: <%p> %p\n", i, n, n->skip[0]);
#endif

        n = n->skip[0];
    }

#endif

    return bk->totalsize;
}


/*
 * return
 *      1   fine
 *      0   strange
 */
int
sk_emptyentity_check(bucket_t *bk)
{
    int ret;
    skiplist_t *list;
    register skiplist_cell_t *x;

    ret = 1;
    list = &bk->cup;
    x = list->header;
    x = x->skip[0];
    while(x!=list->term) {
        if(x->entity == NULL) {
            Error("sk_emptyentity_check: bk# %d, elem %p have empty entityd",
                bk->bid, x);
            ret = 0;
            goto out;
        }
        x = x->skip[0];
    }

out:
    return ret;
}


/*
 * return
 *      1   fine
 *      0   strange
 */
int
sk_entity_check(bucket_t *bk, char *f, int l)
{
    int ret;
    skiplist_t *list;
    register skiplist_cell_t *x;
    register int c, n, tn;

    ret = 1;
    c = 0;
    n = 0;
    tn = 0;
    list = &bk->cup;
    x = list->header;
    x = x->skip[0];
    while(x!=list->term) {
        if(x->entity == NULL) {
            Error(
"sk_entity_check: bk# %d, pos %d, elem %p have empty entity :%s %d",
                bk->bid, c, x, f, l);
            ret = 0;
            n++;
        }
        else
        if(((skcache_entry_t*)x->entity)->issue_time < 0) {
            Error(
"sk_entity_check: bk# %d, pos %d, elem %p have ignore issue_time %d :%s %d",
                bk->bid, c, x, ((skcache_entry_t*)x->entity)->issue_time, f, l);
            ret = 0;
            n++;
        }
        else
        if(((skcache_entry_t*)x->entity)->oid < 0) {
            Error(
"sk_entity_check: bk# %d, pos %d, elem %p have ignore oid# %d :%s %d",
                bk->bid, c, x, ((skcache_entry_t*)x->entity)->oid, f, l);
            ret = 0;
            n++;
        }

        if(((skcache_entry_t*)x->entity)->oid == 0) {
            tn++;
        }

        x = x->skip[0];
        c++;
    }
    if(n>0) {
        Trace("sk_entity_check: bk# %d, found %d ignore entities: %s %d\n",
            bk->bid, n, f, l);
    }
    if(tn>0) {
        Trace("sk_entity_check: bk# %d, found %d temporary entities: %s %d\n",
            bk->bid, tn, f, l);
    }

out:
    return ret;
}






#include "xsk_hash.c"
extern int calc_hash(unsigned char *b, int *h1, int *h2, int *h3, int *h4);


void
dsig_mkhumanstr(u_int32_t *s, char *line)
{
    int i;
    u_int32_t m, v;
    char *p;

    p = line;
    for(i=0;i<DSIG_BIT_WIDTH;i++) {
        m = 1U << i%(SIZE_OF_UINT*8);
        v = s[i/(SIZE_OF_UINT*8)] & m;
        *p++ = (v ? '1' : '0');
    }
    *p = '\0';
}

void
dsig_dump(u_int32_t *s)
{
    char line[DSIG_BIT_WIDTH+2];    /* keep more 2 chars for \n + \0 */

    dsig_mkhumanstr(s, line);
    Trace("%s\n", line);
}

void
dsig_show(u_int32_t *s)
{
    char line[DSIG_BIT_WIDTH+2];    /* keep more 2 chars for \n + \0 */
    int i, c;

    dsig_mkhumanstr(s, line);
    c = 0;
    for(i=0;i<DSIG_BIT_WIDTH;i++) {
        if(DSIG_TMP_ISSET(s, i)) {
            c++;
        }
    }
    Trace("addr %p count %d, bits %s\n", s, c, line);
}


void
dsig_set(u_int32_t *s, int i)
{
    u_int32_t m;
    m = 1UL << i%(SIZE_OF_UINT*8);
    s[i/(SIZE_OF_UINT*8)] |= m;
}

int
dsig_isset(u_int32_t *s, int i)
{
    u_int32_t m;
    m = 1UL << i%(SIZE_OF_UINT*8);
    return (s[i/(SIZE_OF_UINT*8)] & m) ? 1 : 0;
}

int
dsig_allset(u_int32_t *s)
{
    int i;
    u_int32_t u;

    for(i=0;i<DSIG_ELEM_WIDTH;i++) {
        u = (u_int32_t)-1L; /* all bits were set */
        s[i] = u;
    }
}

int
dsig_allclear(u_int32_t *s)
{
    int i;
    u_int32_t u;

    for(i=0;i<DSIG_ELEM_WIDTH;i++) {
        u = (u_int32_t)0UL;
        s[i] = u;
    }
}




int
skcache_entry_show(skcache_entry_t* e, FILE *ofp)
{
    fprintf(ofp,"\
-/-/\n\
entry %p, oid# %d, last# %d, hash %d:%d:%d:%d\n\
name %s\n\
count %d, ref %d, cracked %d, owner %p\n\
issue %d, hot %d, lmd %d, expire %d\n\
totallength %lld\n\
mem record %p, disk record %d\n\
\n",
        e, e->oid, e->last_oid,
        e->h1, e->h2, e->h3, e->h4,
        e->name,
        e->count, e->ref, e->cracked,
        e->owner, 
        e->issue_time, e->hot, e->lmd, e->expire,
        (long long)e->totallength,
        e->m_recorded, e->d_recorded);
    
    return 0;
}


int
skcache_entry_cmp(void *px, void *py)
{
    skcache_entry_t *x = (skcache_entry_t*)px;
    skcache_entry_t *y = (skcache_entry_t*)py;

    if(x->h1==y->h1) {
        if(x->h2==y->h2) {
            if(x->h3==y->h3) {
                return strcmp((const char*)x->name, (const char*)y->name);
            }
            else {
                return x->h3 - y->h3;
            }
        }
        else {
            return x->h2 - y->h2;
        }
    }
    else {
        return x->h1 - y->h1;
    }
}

int
X_skcache_entry_cmp(void *px, void *py)
{
    skcache_entry_t *x = (skcache_entry_t*)px;
    skcache_entry_t *y = (skcache_entry_t*)py;

    if(x->h1==y->h1) {
        if(x->h2==y->h2) {
            return x->h3 - y->h3;
        }
        else {
            return x->h2 - y->h2;
        }
    }
    else {
        return x->h1 - y->h1;
    }
}

char*
skcache_label_hash(void *pp)
{
    skcache_entry_t *p = (skcache_entry_t*)pp;
    static char tmp[L_PATH];

    sprintf(tmp, "%04d:%04d:%04d", p->h1, p->h2, p->h3);

    return tmp;
}

char*
skcache_label_short(void *pp)
{
    skcache_entry_t *p = (skcache_entry_t*)pp;
    static char tmp[L_URL*2];

    sprintf(tmp, "%04d:%04d:%04d %s", p->h1, p->h2, p->h3, p->name);

    return tmp;
}

char*
skcache_label_ref(void *pp)
{
    skcache_entry_t *p = (skcache_entry_t*)pp;
    static char tmp[L_URL*2];

    sprintf(tmp, "%04d:%04d:%04d-%03d", p->h1, p->h2, p->h3, p->ref);

    return tmp;
}

char*
skcache_label4_ref(void *pp)
{
    skcache_entry_t *p = (skcache_entry_t*)pp;
    static char tmp[L_URL*2];

    sprintf(tmp, "%04d:%04d:%04d:%04d-%03d",
        p->h1, p->h2, p->h3, p->h4, p->ref);

    return tmp;
}

char*
skcache_label_long(void *pp)
{
    skcache_entry_t *p = (skcache_entry_t*)pp;
    static char tmp[L_URL*2];

    sprintf(tmp, "%03d:%03d:%03d:%04d=%06d-%03d-0x%08x",
        p->h1, p->h2, p->h3, p->h4, p->oid, p->ref, p->owner);

    return tmp;
}

int
skcache_count_frag_in_entry(skcache_entry_t *en)
{
    int c;
    skcache_frag_t *p;

    c = 0;
    p = en->m_recorded;
    while(p) {
        c++;
        p = p->next;
    }

    return c;
}


int
skcache_dump_entry_frag(char *msg, skcache_entry_t *en)
{
    skcache_frag_t *p;
    off_t frag_sum;
    int u;

    Trace("%s\n", msg);
    Trace("  frags of entry %p, oid# %d,\n", en, en->oid);
    if(en==NULL)
        return -1;

    p = en->m_recorded;
    frag_sum = 0L;
    u = 0;
    while(p) {
        frag_sum += p->len;
        Trace("  %4d: p %p next 0x%08x | len %4d/%d %7ld\n",
            u, p, p->next, p->len, p->size, (long)frag_sum);
        if(p==p->next) {
            Trace("  looped !!!\n");
            break;
        }
        p = p->next;
        u++;
    }
    Trace("  num %d, totalength %lld\n",
        u, (long long)en->totallength);

    return u;
}


off_t
skcache_show_with_HTML(int ofd)
{
    int i;
    off_t wsum, w;
    skiplist_t      *list;
    skiplist_cell_t *cur;
    skcache_entry_t *en;

    wsum = 0;

    w = fdprintf(ofd, "<html>\n");
    wsum += w;



    for(i=0;i<nbucket;i++) {
        w = fdprintf(ofd, "<p>bucket# %d\n", i);
        wsum += w;

        w = fdprintf(ofd, "<listing>\n");
        wsum += w;

        list = &skcache_bucket_base[i].cup;
        cur = list->header->skip[0];

        while(cur && cur!= list->term) {
            en = (skcache_entry_t*)cur->entity;
            w = fdprintf(ofd, " %04d:%04d:%04d %d %d '%s'\n",
                    en->h1, en->h2, en->h3, en->ref, en->count, en->name);
            if(w>0) {
                wsum += w;
            }

            cur = cur->skip[0];
        }

        w = fdprintf(ofd, "</listing>\n");
        wsum += w;

    }


    w = fdprintf(ofd, "<html>\n");
    wsum += w;

    Trace("write %ld bytes\n", wsum);
}


int
skcache_have(int fh, int sh, int th, int ff, char *name,
    int flag_check_rcode, time_t ref_time)
{
    int ret;
    skcache_entry_t tmp;
    skcache_entry_t *stored_pos;
    bucket_t *bk;


    ret = 0;

    bk = &skcache_bucket_base[fh];

    tmp.h1 = fh;
    tmp.h2 = sh;
    tmp.h3 = th;
    tmp.h4 = ff;
    strcpy(tmp.name, name);

    LOCK_BK(bk);

    stored_pos = (skcache_entry_t*) sk_find(&bk->cup, &tmp, skcache_entry_cmp);

    if(stored_pos) {
#if 1
        if(strcmp((const char*)stored_pos->name, (const char*)name)!=0) {
            Error("skcache_have: hash collision %04d:%04d:%04d '%s' v.s. '%s'",
                fh, sh, th, stored_pos->name, name);
            goto out;
        }
#endif

        if(flag_check_rcode) {
        }

        if(ref_time>0) {
        }

        ret = 1;
    }
    else {
        ret = 0;
    }

out:
    UNLOCK_BK(bk);

    return ret;
}


static
long long
alloc_entry_freelist(skcache_entry_t** head, int num)
{
    char *base;
    int esize;
    int i;
    int ret;
    skcache_entry_t *pos;
    skcache_entry_t *last;

    esize = ESIZE;
    base = (char*)MALLOC(esize * num);
    if(base==NULL) {
        return 0;
    }

    for(i=0; i<num; i++) {
        pos = (skcache_entry_t*)(base + esize*i);

        pos->magic[0] = MAGIC_0;
        pos->magic[1] = MAGIC_1;

        pos->oid = -1;
        pos->ref = -1;
        pos->h1 = -1;
        pos->h2 = -1;
        pos->h3 = -1;
        pos->name[0] = '\0';
        pos->m_recorded = NULL;
        pos->d_recorded = -1;

        pos->next = (skcache_entry_t*)((char*)pos + esize);
    }

    *head = (skcache_entry_t*)base;
    pos->next = NULL;

#if 0

    Trace("entry// head %p, num %d\n", head, num);
    Trace("esize %d, base %p\n", esize, base);

    if(num<100) {
        i = 0;
        pos = *head;
        while(pos) {
            Trace(" %2d: pos %p, name '%s', m_recorded %p ->next %p\n",
                i, pos, pos->name, pos->m_recorded, pos->next);

            pos = pos->next;
            i++;
        }
    }

#endif

    return esize*num;
}

int
skcache_frag_freelist_check(skcache_frag_t* base)
{
    int i;
    skcache_frag_t *pos;
    int sumsz, sumlen;
    
    pos = base;
    i = 0;
    sumsz = sumlen = 0;
    while(pos) {
        sumsz  += pos->size;
        sumlen += pos->len;
        pos = pos->next;
        i++;
    }
    Trace("skcache_frag_freelist_check: base %p, count %d, length sum %d, av. %.2f %d // size sum %d av. %.2f %d\n",
        base, i,
        sumlen, ((double)sumlen)/i , sumlen%i, 
        sumsz, ((double)sumsz)/i , sumsz%i );

    return i;
}

static
long long
alloc_frag_freelist_single_chunk(skcache_frag_t** head, int num, int sz)
{
    char *base;
    int esize;
    int i;
    skcache_frag_t *pos;
    skcache_frag_t *last;
    int sum;

    esize = sizeof(skcache_frag_t)+sz;
    base = (char*)MALLOC(esize * num);
    if(base==NULL) {
        return 0;
    }

    for(i=0; i<num; i++) {
        pos = (skcache_frag_t*)(base + esize*i);
        pos->size = sz;
        pos->len = 0;
        pos->ent = (char*)((char*)pos+sizeof(skcache_frag_t));

        pos->next = (skcache_frag_t*)((char*)pos+esize);
    }
    *head = (skcache_frag_t*)base;
    pos->next = NULL;

    skcache_frag_freelist_check(*head);

#if 1
    Trace("frag// head %p, num %d, sz %d // total %lld\n",
        head, num, sz, (long long)esize*num);
    Trace("esize %d, base %p\n", esize, base);
#endif

    return esize*num;
}


/*
 * alloc_frag_freelist_double_chunk
 */
static
long long
alloc_frag_freelist(skcache_frag_t** head, int num, int sz)
{
    char *base;
    char *wing;
    int esize;
    int i;
    skcache_frag_t *pos;
    skcache_frag_t *last;
    int sum;
    long long ret;
    int psize;

    ret = 0LL;

#if 0
    if(sz%FRAG_ALIGN!=0) {
        Error("frag size %d is not alignmented by %d", sz, FRAG_ALIGN);
        sz = (sz<FRAG_ALIGN) ? FRAG_ALIGN: ((sz+FRAG_ALIGN-1)/FRAG_ALIGN)*FRAG_ALIGN;
    }
#endif
#if 1
    Trace("using frag size is %d (= %d * %d)\n", sz, FRAG_ALIGN, sz/FRAG_ALIGN);
#endif


    esize = sizeof(skcache_frag_t);
    base = (char*)MALLOC(sizeof(skcache_frag_t) * num);
    if(base==NULL) {
        return 0;
    }
    ret += sizeof(skcache_frag_t) * num;
    mlist_add_elem("frag base", (void*)base, sizeof(skcache_frag_t)*num);

#ifdef HAVE_GETPAGESIZE
    psize = getpagesize();
#else
    psize = 4096;
#endif
    wing = (char*)MALLOC(sz * num + psize);
    if(wing==NULL) {
        return 0;
    }
    ret += sz * num + psize;
    mlist_add_elem("frag body", (void*)wing, sz * num + psize);

#if 1
    Trace("  allocate %dB, %.2fM\n", sz*num+psize, fdM(sz*num+psize));
#endif
#if 1
    Trace("  wing (before alignment) %p\n", wing);
#endif

    wing = wing + psize - ((int)wing%psize);

#if 1
    Trace("  wing (after  alignment) %p , with page-size %d\n", wing, psize);
#endif


    for(i=0; i<num; i++) {
        pos = (skcache_frag_t*)(base + esize*i);
        pos->size = sz;
        pos->len = 0;
#if 0
        pos->ent = (char*)((char*)pos+sizeof(skcache_frag_t));
#endif
        pos->ent = (char*)(wing+sz*i);

        pos->next = (skcache_frag_t*)((char*)pos+esize);
    }
    *head = (skcache_frag_t*)base;
    pos->next = NULL;

    skcache_frag_freelist_check(*head);

#if 1
    Trace("frag// head %p, num %d, sz %d // total %lld\n",
        head, num, sz, (long long)esize*num);
    Trace("esize %d, base %p\n", esize, base);
    Trace("allocate %lld bytes\n", ret);
#endif

    return ret;
}

int
echk_count=0;

int
skcache_entry_check(skcache_entry_t* e, char*f, int l)
{
    int ret;
    int cur_count;

    ret = 0;

    if(e==NULL) {
        Error("echk: %s %d : empty entry", f, l);
        ret = -1;
        goto out;
    }
#ifdef CHECK_MAGIC
    if(e->magic[0]!=MAGIC_0 || e->magic[1]!=MAGIC_1) {
        Error("echk: %s %d : entry %p, oid# %d, ignore magic",
            f l, e, e->oid);
        ret = 2;
        goto out;
    }
#endif
#if 0
    if(e->ref<0) {
        Error("echk: %s %d : entry %p, oid# %d, owner %p, neg ref %d",
            f, l, e, e->owner, e->oid, e->ref);
        ret = 3;
        goto out;
    }
#endif
#if 1
    if(e->ref<0) {
        e->ref = 0;
    }
#endif

out:
#if 0
    if(ret) {
        pthread_mutex_lock(&stat_lock);
        echk_count++;
        cur_count = echk_count;
        pthread_mutex_unlock(&stat_lock);
        if(cur_count>3) {
            skcache_entry_show(e, stdout);
            exit(0);
        }
    }
#endif

    return ret;
}

int skcache_push_idle_frag(bucket_t *bk, skcache_frag_t *xf);
int RAW_skcache_push_idle_frag(bucket_t *bk, skcache_frag_t *xf);
static int multi_frag_lock();

#if 1
#define MFRAG_LOCK()    (multi_frag_lock() ? Error("cannot lock mutex %s %d", __FILE__, __LINE__), 1 : 0 )
#else
#define MFRAG_LOCK()    multi_frag_lock()
#endif


int
skcache_entry_purge(bucket_t *bk, skcache_entry_t *e)
{
    skcache_frag_t *cur, *pre;
    int chk;

    chk = MFRAG_LOCK();
    if(chk) {
        Error("skcache_entry_purge: fail frag mutex lock");
        return -1;
    }

    if(e->m_recorded) {
        /* remove */
        cur = e->m_recorded;
        while(cur) {
            pre = cur->next;

            RAW_skcache_push_idle_frag(bk, cur);
            bk->own_frag--;
            if(pre==cur) {
                Error(
                  "skcache_entry_purge: bk %p, entry %p, frag %p is looped",
                    bk, e, cur);
                break;
            }

            cur = pre;
        }

        e->m_recorded = NULL;
    }

    MUTEX_UNLOCK(&frag_lock);

    return 0;
}



int
skcache_entry_delete(bucket_t *bk, skcache_entry_t *e)
{
    void *chkv;
    int chk;
    int ret;

    if(ret = skcache_entry_check(e, __FILE__, __LINE__)) {
        Trace("skcache_entry_delete: bk# %d, entry %p is ignore, but delete\n",
            bk->bid, e);
#if 1
        goto out;
#endif
    }
    ret = 0;

#if 0
    Trace("skcache_entry_delete: bk# %d %p, entry %p, oid# %d, ref %d\n",
        bk->bid, bk, e, e->oid, e->ref);
#endif

 assert(sk_emptyentity_check(bk));
    chk = skcache_entry_purge(bk, e);
 assert(sk_emptyentity_check(bk));
    if(chk) {
        Error("skcache_entry_delete: fail skcache_entry_purge() bk# %d, entry %x",
            bk->bid, e);
        ret = 1;
    }
    else {

#if 0
    chk = sk_entity_check(bk, __FILE__, __LINE__);
#endif
 assert(sk_emptyentity_check(bk));
        chkv = sk_delete(&bk->cup, e,
            &bk->elem_freelist, skcache_entry_cmp, NULL, 0);
 assert(sk_emptyentity_check(bk));
        if(chkv==NULL) {
            Trace("skcache_entry_delete: bk# %d, entry %p is not found\n",
                bk->bid, e);

            ret = 2;

            chkv = sk_find(&bk->cup, e, skcache_entry_cmp);
            Trace("  sk_find() returns %x\n", chkv);

        }
        else {
            chk = skcache_push_idle_entry(bk, e);
#if 1
            if(chk) {
                Trace("skcache_entry_delete: chk %d\n", chk);
                ret = 3;
            }
#endif
        }

#if 0
    chk = sk_entity_check(bk, __FILE__, __LINE__);
#endif
    }

out:
    return ret;
}







#ifdef NO_VS
#ifdef VS
#undef VS
#endif
#endif



/*
 * return code is describe in comment in this routine.
 */
int
skcache_dicision_w_th(bucket_t *bk, skcache_entry_t *p, int force,
    time_t now, time_t min, time_t th, time_t max)
{
    int elap, idle;
    int action;
    int chk;
    char *dmsg;
    char line[L_BUF];
    char tmp[L_BUF];

#if 0
    Trace(
"_dicision_w_th: bk# %d, p 0x%x, force %d // min,thre,max %d,%d,%d\n",
        bk->bid, p, force, min, th, max);
#endif

    action = ACTION_NOP;

    /*
     * calc elap time from barth
     */
    if(p->issue_time > 0 ) {
        elap = now - p->issue_time;
    }
    else {
        elap = -1;
    }



    if(p->oid == 0 ) {
        if(p->owner) {
#if 0
            Trace(
"_dc: temporary  p %p %d oid# %d, ref %d, count %d, size %6ld, owner %p\n",
    p, p->issue_time, p->oid, p->ref, p->count, (long)p->totallength, p->owner);
#endif
            action = ACTION_NOP;
            goto skip;
        }
        else {
#if 0
            Trace(
"_dc: losted tmp p %p %d oid# %d, ref %d, count %d, size %6ld, owner %p\n",
    p, p->issue_time, p->oid, p->ref, p->count, (long)p->totallength, p->owner);
#endif
            action = ACTION_PURGE;
            goto skip;
        }
    }



    /*
     * calc idle time from last action
     */
    if(p->hot > 0 ) {
        idle = now - p->hot;
    }
    else {
#if 0
        Trace("empty hot\n");
#endif
#if 0
        idle = MIN(th/2, elap);
#endif
        idle = -1;
    }


#ifdef VS
    sprintf(tmp, "bk# %d, oid# %d,", p->h1, p->oid);
    sprintf(line,
        "%-24s %p: %7lld %c%c%c %p %8d, idle %4d, elap %5d: ",
        tmp, p, (long long)p->totallength,
        (p->ref > 0)                ? 'R' : '-',
        (p->cracked > 0)            ? 'C' : '-',
        (p->swapout_candidate > 0)  ? 'S' : '-',
        p->m_recorded, p->d_recorded,
        idle, elap);
#endif

#if 0
    sprintf(tmp, "bk# %d, oid# %d,", p->h1, p->oid);
    sprintf(line,
        " %-32s %p: %7lld %c%c%c %p %3d, idle %4d, elap %5d: ",
        tmp, p, (long long)p->totallength,
        (p->ref > 0)                ? 'R' : '-',
        (p->cracked > 0)            ? 'C' : '-',
        (p->swapout_candidate > 0)  ? 'S' : '-',
        p->m_recorded, p->d_recorded,
        idle, elap);
#endif

#if 0
#ifdef VS
    if(p->m_recorded) {
    sprintf(line,
        " %p:: %8d %7ld %c 0x%08x %3d, c %d, f %2d, idle %4d, elap %5d: ",
        p, p->oid, (long)p->totallength,
        (p->ref > 0) ? 'R' : '-',
        p->m_recorded, p->d_recorded, p->cracked, p->count,
        idle, elap);
    }
    else {
    sprintf(line,
        " %p:: %8d %7ld %c %5s %8d, c %d, f %2d, idle %4d, elap %5d: ",
        p, p->oid, (long)p->totallength,
        (p->ref > 0) ? 'R' : ' ',
        "-", p->d_recorded, p->cracked, p->count,
        idle, elap);
    }
#endif
#endif

#if 0
    Trace("lmd %d now %d diff %d ", p->lmd, now, now - p->lmd);
#endif

#ifdef VS
    dmsg = "unknow  ";
#endif


    /* action
        3   swapout + purge
        2   purge
        1   swapout
        0   no action
     */


     /***
      *** black and white zone
      ***/
    if(p->ref>0) {
#ifdef VS
    dmsg = "refered ";
#endif
        action = ACTION_NOP;
    }
    else
#if 0
    if(p->swapout_candidate>0) {
#ifdef VS
    dmsg = "marked  ";
#endif
        action = ACTION_NOP;
    }
    else
    if(p->swapout_candidate<0) {
#ifdef VS
    dmsg = "ignore3 ";
#endif
        action = ACTION_PURGE;
    }
    else
#endif
    if(elap<0) {
#ifdef VS
    dmsg = "junk    ";
#endif
        action = ACTION_PURGE;
    }
    else
    if(p->cracked) {
#ifdef VS
    dmsg = "cracked ";
#endif
        action = ACTION_PURGE;
    }
    else
    if(p->m_recorded==NULL) {
#ifdef VS
    dmsg = "empty-b ";
#endif
        action = ACTION_PURGE;
    }
    else
    if(p->totallength<=(off_t)0) {
#ifdef VS
    dmsg = "empty-c ";
#endif
        action = ACTION_PURGE;
    }
    else
    if(p->expire>0 && p->expire <= now) {
        /* MUST expire */
#ifdef VS
    dmsg = "expireM ";
#endif
        action = ACTION_PURGE;              /* don't swapout */
    }
#if 0
    else 
    if(p->lmd>0 && now - p->lmd > max) {
        /* maybe old, SHOULD expire */
#ifdef VS
    dmsg = "expireS ";
#endif
        action = ACTION_PURGE;
    }
#endif

    /***
     *** gray // dynamic zone
     ***/

    else {
        if(elap>th) {
            if(p->d_recorded>=0 && p->d_recorded==p->oid) {
#ifdef VS
    dmsg = "stored  ";
#endif
                action = ACTION_PURGE;
            }
            else
            if(p->totallength>OFILE_BLKSIZE*OFILE_NBLK_OBJ) {
#ifdef VS
    dmsg = "too big ";
#endif
                action = ACTION_PURGE;
            }
            else {
#ifdef VS
    dmsg = "swapout ";
#endif
                action = ACTION_SWAPOUT_PURGE;
            }
        }
#if 1
        else
        if(force) {
#ifdef VS
    dmsg = "f-swapo ";
#endif
            action = ACTION_SWAPOUT_PURGE;
        }
#endif
        else {
#ifdef VS
    dmsg = "__keep_ ";
#endif
            action = ACTION_NOP;
        }

    }

#ifdef VS
Trace("%s %s|\n", line, dmsg);
#endif

skip:
    return action;

}

int
skcache_dicision(bucket_t *bk, skcache_entry_t *p, int force)
{
    time_t now;

    now = time(NULL);

    return skcache_dicision_w_th(bk, p, force,
        now,
        skcache_mem_keep_min, skcache_mem_keep_cur, skcache_entry_keep_cur);
}







int check_bkinfo();


skcache_entry_t*
skcache_pull_new_entry(bucket_t *bk)
{
    int chk;
    skcache_entry_t* newone;

#if 0
#ifndef NDEBUG
    Trace("new_entry: bk# %d, %p, use %d, free %d, cup.num %d\n",
        bk->bid, bk, bk->entry_use, bk->entry_free, bk->cup.num);
#endif
#endif

    if(bk->entry_freelist == NULL) { 
        Error("new_entry: no free item");
        return NULL;
    }

    newone = bk->entry_freelist;
    bk->entry_freelist = newone->next;
    bk->entry_use++;
    bk->entry_free--;

    memset(newone, 0, ESIZE); /* clear all field */
    newone->magic[0] = MAGIC_0;
    newone->magic[1] = MAGIC_1;
    newone->oid = 0;
    newone->h1 = newone->h2 = newone->h3 = newone->h4 = -1;
    newone->ref = 0;
    newone->count = 0;
    newone->totallength = (off_t)0;
    newone->issue_time = -1;
    newone->swapout_candidate = 0;
    newone->d_recorded = 0;
    newone->m_recorded = NULL;

    return newone;
}

int
skcache_push_idle_entry(bucket_t *bk, skcache_entry_t *xen)
{
    int last;
#if 0
#ifndef NDEBUG
    Trace("idle_entry: bk# %d, %p, use %d, free %d, cup.num %d, target %p\n",
        bk->bid, bk, bk->entry_use, bk->entry_free, bk->cup.num, xen);
#endif
#endif

    last = xen->oid;

    memset(xen, 0, ESIZE);
    xen->last_oid = last;
    xen->h1 = xen->h2 = xen->h3 = xen->h4 = -1;
    xen->ref = -1;
    xen->issue_time = -1;
    xen->swapout_candidate = -1;

    xen->next = bk->entry_freelist;
    bk->entry_freelist = xen;
    bk->entry_use--;
    bk->entry_free++;

    return 0;
}


/* success 0, fail 1 */
static
int
multi_frag_lock()
{
    int c;
    int chk;

#if 0
    /* for debug */
    if(frag_use<=-50) {
        fflush(stdout);
        Error("multi_frag_lock: frag under flow, use %d, free %d",
            frag_use, frag_free);
        fflush(stderr);
        fflush(stdout);
        exit(4);
    }
#endif

    c = FRAG_NTRY;
try_lock:
    c--;
    chk = MUTEX_TRYLOCK(&frag_lock);
    if(chk==EBUSY) {
        if(c>0) {
#if 0
    Trace("multi_frag_lock: free list is busy, retry %d : u %d, f %d\n",
                c, frag_use, frag_free); 
#endif
            msleep(3);
            goto try_lock;
        }
        Error("multi_frag_lock: free list is busy <%d times>, u %d, f %d",
            c, frag_use, frag_free);
        FCHK();
        return 1;
    }
    else
    if(chk) {
        Error("multi_frag_lock: unexpected error (%d)", chk);
        FCHK();
        return 3;
    }

    return 0;
}


skcache_frag_t*
skcache_pull_new_frag(bucket_t *bk)
{
    skcache_frag_t* cur;
    int chk;
    int c;

#if 0
    Trace("new_frag: use %d, free %d\n",
        frag_use, frag_free);
#endif

#if 1
    if(frag_use<0) {
        Error("new_frag: strange the number of used frag is negative %d",
            frag_use);
        return NULL;
    }
#endif

    if(MFRAG_LOCK()) {
        if(bk)
            Error("new_frag: cannot lock fraglist (bk# %d)", bk->bid);
        else
            Error("new_frag: cannot lock fraglist (unknown bk)");
        
        FCHK();
        return NULL;
    }

    if(z_frag_freelist == NULL || frag_free <= 0 ) { 
        Error("new_frag: no free item, use %d %.2f%%, pointer %p",
            frag_use, (double)100.0*frag_use/(frag_use+frag_free),
            z_frag_freelist);

        MUTEX_UNLOCK(&frag_lock);

        FCHK();
        return NULL;
    }

    cur = z_frag_freelist;
    z_frag_freelist = cur->next;
    frag_use++;
    frag_free--;
    cur->next = NULL;
    cur->len  = 0;

#if 0
    if(frag_free<=1) {
        Error("new_frag: left last one frag %p %d",
            z_frag_freelist, frag_free);
    }
#endif

    MUTEX_UNLOCK(&frag_lock);

    FCHK();
    return cur;
}

int
RAW_skcache_push_idle_frag(bucket_t *bk, skcache_frag_t *xf)
{
    xf->next = z_frag_freelist;
    z_frag_freelist = xf;
    frag_use--;
    frag_free++;

    return 0;
}

int
skcache_push_idle_frag(bucket_t *bk, skcache_frag_t *xf)
{
    int chk;
    int c;

#if 0
    Trace("idle_frag: target %p, use %d, free %d\n",
        xf, frag_use, frag_free);
#endif
    if(xf==NULL) {
        return -13;
    }
    if(MFRAG_LOCK()) {
        return -1;
    }

    chk = RAW_skcache_push_idle_frag(bk, xf);

    MUTEX_UNLOCK(&frag_lock);
    
    FCHK();
    return 0;
}



/*
 * re-calcurate keeptime threshould
 * argument:
 *      delta_t [msec]
 */
int
skcache_update_keeptime(int grow, int delta_t)
{
    time_t now;
    int last, keep;
    double cratio; int cuse;
    double hratio; int huse;
    int rapid_flag;
    int i;
    bucket_t *bk;
    int chk;
    char *msg;
    int elap;
    extern time_t boot_date;

    now = time(NULL);
    rapid_flag = 0;
    elap = now - boot_date;

    MUTEX_LOCK(&config_lock);
    hratio = sim_hist_average_d(&fratio_hist);
    huse   = (int)(hratio*frag_all/100);
    cratio = 100.0*frag_use/frag_all;
    last = keep = skcache_mem_keep_cur;

    if(skcache_keeptime_token>0) {
        skcache_keeptime_token--;
        goto skip_update;
    }

#if 0
    Trace("KT keeptime cur %d; frag cur %d %.2f%%, hist %d %.2f%%; water all/hh/h/l %d/%d/%d/%d\n",
        skcache_mem_keep_cur, 
        frag_use, cratio, 
        huse, hratio, 
        frag_all, frag_hhwater, frag_hwater, frag_lwater);
#endif


    if(huse>=frag_hhwater) {
            msg = "be-quoter";
            skcache_mem_keep_cur = keep/4;
            skcache_keeptime_token = skcache_keeptime_token_init*2;
    }
    else
    if(huse>=frag_hwater) {
        if(frag_use>=frag_hhwater) {
            msg = "be-half";
            skcache_mem_keep_cur = keep/2;
            skcache_keeptime_token = skcache_keeptime_token_init*2/3;
        }
        else {
            msg = "rapid-down";
            skcache_mem_keep_cur = 15*keep/20;
            skcache_keeptime_token = skcache_keeptime_token_init;
        }
    }
    else
    if(huse>frag_lwater) {
        if(frag_use>=frag_hwater) {
            msg = "slow-down";
            skcache_mem_keep_cur = 18*keep/20;
            skcache_keeptime_token = skcache_keeptime_token_init;
        }
        else
        if(frag_use<frag_lwater) {
            msg = "slow-up2";
            skcache_mem_keep_cur += skcache_mem_keep_max/400;   /* 0.25% */
            skcache_keeptime_token = skcache_keeptime_token_init*2;
        }
        else {
            msg = "no-change";
            skcache_keeptime_token = skcache_keeptime_token_init;
        }
    }
    else {
        if(frag_use>=frag_lwater) {
            msg = "slow-up";
            skcache_mem_keep_cur += skcache_mem_keep_max/400;   /* 0.25% */
            skcache_keeptime_token = skcache_keeptime_token_init*2;
        }
        else {
            msg = "up";
            if(grow) {
#if 0
            skcache_mem_keep_cur += skcache_mem_keep_max/50;    /* 2% */
#endif
            skcache_mem_keep_cur += skcache_mem_keep_max/200;   /* 0.5% */
            skcache_keeptime_token = skcache_keeptime_token_init*3;
            }
        }
    }

    /* fix out range values */
    if(skcache_mem_keep_cur<skcache_mem_keep_min) {
        Trace("  underflow min %d\n", skcache_mem_keep_min);
        skcache_mem_keep_cur = skcache_mem_keep_min;
    }
    if(skcache_mem_keep_cur>skcache_mem_keep_max) {
        Trace("  overflow  max %d\n", skcache_mem_keep_max);
        skcache_mem_keep_cur = skcache_mem_keep_max;
    }

#if 0
    Trace("KT updating %d frag %.2f%% (%d) %s thre. %d -> %d\n",
        now, hratio, skcache_keeptime_token, msg, last, skcache_mem_keep_cur);
#endif

#if 0
    Trace("KT %d %d %d; keeptime cur %d; frag cur %d %.2f%%, hist %d %.2f%%; water all/hh/h/l %d/%d/%d/%d :%s -> %d\n",
        elap, grow, delta_t,
        last, 
        frag_use, cratio, 
        huse, hratio, 
        frag_all, frag_hhwater, frag_hwater, frag_lwater,
        msg, skcache_mem_keep_cur);
#endif

#if 1
    Trace("KT %d %d %d; keeptime cur %d; frag cur %d %.2f%%, hist %d %.2f%%; %s -> %d : rest token %d\n",
        elap, grow, delta_t,
        last, 
        frag_use, cratio, 
        huse, hratio, 
        msg, skcache_mem_keep_cur, skcache_keeptime_token);
#endif

skip_update:
    MUTEX_UNLOCK(&config_lock);

    if(rapid_flag) {
        for(i=0;i<radix_h1;i++) {
            bk = &skcache_bucket_base[i];

            chk = MUTEX_TRYLOCK(&bk->mutex);
            if(chk==EBUSY) {
            }
            else {
                bk->gc_push++;
                MUTEX_UNLOCK(&bk->mutex);
            }

        }
    }

    return skcache_mem_keep_cur;
}








long long
skcache_calc_cacheareas(
    int *xbk_num, int *xsk_level, int *xsk_len, int *xffl_len, int *xffl_size)
{
    long long r;
    int esk;
    int tsk;
    int ten;
    int tfl;
    long long ebk;
    long long tbk;

#if 0
    Trace("skcache_calc_cacheareas:\n");
    Trace("  bucket:\n");
#endif

    r = (((sizeof(void*)+sizeof(skiplist_cell_t**)
            + sizeof(skiplist_cell_t*)*(*xsk_level+1))
        + ESIZE)* (*xsk_len)
        + ESIZE
        + (sizeof(skcache_frag_t) + *xffl_size) * (*xffl_len))*(*xbk_num);

    esk = sizeof(void*)+sizeof(skiplist_cell_t**)
            +sizeof(skiplist_cell_t*)*(*xsk_level+1);
    tsk = esk * *xsk_len;
#if 0
    Trace("    skiplist:\n");
    Trace("      level %d\n", *xsk_level);
    Trace("      -> element size of skiplist %ld\n", esk);
    Trace("      length of skiplist %d\n", *xsk_len);
#endif

    ten = ESIZE*(*xsk_len+1);
    tfl = (sizeof(skcache_frag_t)+ *xffl_size)*(*xffl_len);
    ebk = (long long)(sizeof(bucket_t)+tsk+ten+tfl);
#if 0
    Trace("    -> size of skiplist            %12ld %4ldM+%07ld %7.2f%%\n",
        tsk, (long)dM(tsk), (long)mM(tsk), ((double)tsk*100)/ebk);
    Trace("    size of free-list for entities %12ld %4ldM+%07ld %7.2f%%\n",
        ten, (long)dM(ten), (long)mM(ten), ((double)ten*100)/ebk);
    Trace("    size of fragment free-list     %12ld %4ldM+%07ld %7.2f%%\n",
        tfl, (long)dM(tfl), (long)mM(tfl), ((double)tfl*100)/ebk);
    Trace("  -> size of bucket                %12lld %4ldM+%07ld\n",
        ebk, (long)dM(ebk), (long)mM(ebk));
#endif


    tbk = ebk * *xbk_num;

#if 0
    Trace("  number of bucket %ld\n", *xbk_num);
    Trace("-> total size of all buckets       %12lld %4ldM+%07ld\n",
        tbk, (long)dM(tbk), (long)mM(tbk));
    Trace("-> total number of entries         %12ld\n", *xsk_len * *xbk_num);
    Trace("-> total number of frags           %12ld\n", *xffl_len * *xbk_num);
#endif

    return r;
}

int
skcache_calc_param(long long want,
    int *xbk_num, int *xsk_level, int *xffl_size, double r,
    int *xsk_len, int *xffl_num)
{
    long long bksize;
    long a, b, c;

#if 0
    Trace("skcache_calc_param: want %lld\n", want);
#endif

#if 0
    Trace("    specified xbk_num %d, xsk_level %d, xffl_size %d r %.2f\n",
        *xbk_num, *xsk_level, *xffl_size, r);
#endif

    if(*xffl_size<0) {
        *xffl_size = 1460;
        Trace("  update frag size as %d\n", *xffl_size);
    }
#if 0
    if(*xffl_size%FRAG_ALIGN!=0) {
        Error("frag size %d is not alignmented by %d", *xffl_size, FRAG_ALIGN);
        *xffl_size = (*xffl_size<FRAG_ALIGN) ?
            FRAG_ALIGN: ((*xffl_size+FRAG_ALIGN-1)/FRAG_ALIGN)*FRAG_ALIGN;
        Trace("  update frag size as %d\n", *xffl_size);
    }
#endif

    bksize = want/(*xbk_num) - ESIZE;

 if(*xsk_len<0) {

    a = sizeof(void*)+sizeof(skiplist_cell_t**)
            +sizeof(skiplist_cell_t*)*(*xsk_level+1);
    b = ESIZE;
    c = sizeof(skcache_frag_t)+ *xffl_size;

#if 0
    Trace("    bucket size %lld\n", bksize);
    Trace("    a %ld\n", a);
    Trace("    b %ld\n", b);
    Trace("    c %ld\n", c);
#endif

    *xsk_len = (long)
        (bksize - b) / (a + b + (long)(r*c));
    *xffl_num = (long)(r * *xsk_len) * *xbk_num;

  }
  else {
    *xffl_num = -1;
  }


#if 0
    Trace("    solved    xsk_len %d, xffl_num %d\n",
        *xsk_len, *xffl_num);
#endif


    return *xsk_len;
}

/*
 * initialized cache parameters and areas
 */
long long
skcache_alloc_cacheareas(time_t xmin_date,
    int xbk_num, int xsk_level, int xsk_len, int xffl_num, int xffl_size)
{
    int i,j;
    int chk;
    long long got;
    long long sum, tmp_sum;
    long long all;
    bucket_t *bk;
    time_t now;

    long long tmp;


#if 1
    Trace("skcache_alloc_cacheareas: min_date %d\n\
    bucket: number %d, level %d, length %d\n\
    frag: length %d, size %d\n",
        xmin_date, xbk_num, xsk_level, xsk_len, xffl_num, xffl_size);
#endif

    if(xmin_date>0) {
        min_date = xmin_date;
    }
    else {
#if 0
        min_date = time(NULL);
        Trace("set min_date as %d (current date/time)\n", min_date);
#endif
        Trace("set min_date as negative it is mean, use stored objects before boot\n");
    }

    now = time(NULL);

    sum = 0L;



    /*
     * NIL
     */
    NIL = (skcache_entry_t*)MALLOC(sizeof(bucket_t));
    NIL->h1 = INT_MAX;
    NIL->h2 = INT_MAX;
    NIL->h3 = INT_MAX;

    sum += sizeof(bucket_t);
    mlist_add_elem("NIL", (void*)NIL, sizeof(bucket_t));

#if 1
    /* XXX */
    nbucket = xbk_num;
#endif

#if 0
    Trace("require_radix_h2 %d, require_radix_h3 %d\n",
            require_radix_h2, require_radix_h3);
#endif
    
    /*
     * cache entries
     */
    radix_h1 = nbucket;
    radix_h2 = require_radix_h2;
    radix_h3 = require_radix_h3;
    radix_h4 = 7537;                /* XXX */


    if(radix_h2>RADIX_H2_MAX) {
        Error("radix_h2 (%d) is too large (max %d)\n",
            radix_h2, RADIX_H2_MAX);
    }
    Trace("\n\
Hashes:\n\
  Radixes               %d, %d, %d, %d\n\
\n",
        radix_h1, radix_h2, radix_h3, radix_h4);



    dir_num = radix_h1*radix_h2;

    if(dir_num<0) {
        Trace("dir_num %d is stange\n", dir_num);
        exit(4);
    }


    skcache_disk_setup();


    dir_lock = (int*)MALLOC(sizeof(int)*dir_num);
    sum += sizeof(int)*dir_num;
    mlist_add_elem("dir_lock",   (void*)dir_lock, sizeof(int)*dir_num);

#if 1
    Trace("\n\
Signatures for directories:\n\
  SIZE_OF_UINT        %d\n\
  DSIG_BIT_WIDTH      %d\n\
  DSIG_VALUE_RANGE    %d\n\
  DSIG_VALUE_RES      %d\n\
  DSIG_ELEM_WIDTH     %d\n\
  DSIG_BYTE_WIDTH     %d\n",
        SIZE_OF_UINT,
        DSIG_BIT_WIDTH, DSIG_VALUE_RANGE, DSIG_VALUE_RES,
        DSIG_ELEM_WIDTH, DSIG_BYTE_WIDTH);
    Trace("\n");
#endif
#if 0
    Trace(" require_total_memory_size %lld\n",
        (long long)require_total_memory_size); 
#endif

    dir_sig = (u_int32_t*)MALLOC(sizeof(u_int32_t)*dir_num*DSIG_ELEM_WIDTH);
    sum += sizeof(u_int32_t)*dir_num*DSIG_ELEM_WIDTH;
    mlist_add_elem("dir_sig",
        (void*)dir_sig, sizeof(u_int32_t)*dir_num*DSIG_ELEM_WIDTH);

    dir_sz  = (off_t*)MALLOC(sizeof(off_t)*dir_num);
    sum += sizeof(off_t)*dir_num;
    mlist_add_elem("dir_sz",    (void*)dir_sz, sizeof(off_t)*dir_num);

    dir_count = (int*)MALLOC(sizeof(int)*dir_num);
    sum += sizeof(int)*dir_num;
    mlist_add_elem("dir_count", (void*)dir_count, sizeof(int)*dir_num);

    dir_time = (int*)MALLOC(sizeof(int)*dir_num);
    sum += sizeof(int)*dir_num;
    mlist_add_elem("dir_time",  (void*)dir_time, sizeof(int)*dir_num);

    dir_lack = (int*)MALLOC(sizeof(int)*dir_num);
    sum += sizeof(int)*dir_num;
    mlist_add_elem("dir_lack",  (void*)dir_time, sizeof(int)*dir_num);


    for(i=0;i<dir_num;i++) {
#if 0
        dir_sig[i]      = (int64_t)-1L;
#endif
        DSIG_ALLSET(i);
        dir_sz[i]       = (off_t)-1L;
        dir_count[i]    = -1;
        dir_time[i]     = now - rand()%skcache_signature_refresh;
        dir_lock[i]     = 0;
        dir_lack[i]     = 0;
    }



#if 0
    Trace(" require_total_memory_size %lld\n",
        (long long)require_total_memory_size); 
#endif



    skcache_bucket_base = (bucket_t*)MALLOC(sizeof(bucket_t)*xbk_num);
    if(skcache_bucket_base==NULL) {
        Error("not enough memory for skcache_bucket_base");
        nbucket = 0;
        return -1;
    }

    sum += sizeof(bucket_t)*xbk_num;
    mlist_add_elem("bucket bases",
        (void*)skcache_bucket_base, sizeof(bucket_t)*xbk_num);

    tmp_sum = 0LL;

    for(i=0;i<nbucket;i++) {
#if 0
Trace("bk# %d %p\n", i, skcache_bucket_base[i]);
#endif

        chk = bk_new(&skcache_bucket_base[i], xsk_level, xsk_len, NIL);
        if(chk==0) {
            Error("initialize error in bucket# %d", i);
            return -2;
        }
        tmp_sum += chk;

        bk = &skcache_bucket_base[i];
        bk->bid = i;
        bk->own = '.';

        bk->fs = FSSOLV(i);

        got = alloc_entry_freelist(&bk->entry_freelist, xsk_len+1);
        if(got<=0) {
            Error("initialize error in object entry elem_freelist# %d", i);
            return -2;
        }
        else {
            tmp_sum += got;
        }

        bk->entry_all  = xsk_len+1;
        bk->entry_use  = 0;
        bk->entry_free = xsk_len+1;

        bk->entry_hhwater = (int) (require_entry_hhwater * bk->entry_all/100);
        bk->entry_hwater  = (int) (require_entry_hwater * bk->entry_all/100);
        bk->entry_lwater  = (int) (require_entry_lwater * bk->entry_all/100);


        bk->own_frag = 0;

        bk->gc_push = 0; 
        bk->gc_rest = 0;
#if 0
        bk->gc_last = -1;
        bk->gc_checkpoint = 
            now - skcache_mem_keep_cur/4 + (rand()% (skcache_mem_keep_cur/2));
#endif
        bk->gc_last = now;
        bk->gc_checkpoint = 
            now - (rand()%(95*skcache_mem_keep_cur/100));
        bk->gc_ntarget = 0;

        bk->dirpos = rand()%radix_h2;
        bk->dirty2 = -1;
        bk->dirty3 = -1;

    }

#if 1
    Trace("Object Cache:\n\
 entries:\n\
  the number of entries %d (/bucket)\n\
  watermark             hhigh %d (%.2f%%), high %d (%.2f%%), low %d (%.2f%%)\n",
            bk->entry_all,
            bk->entry_hhwater,  ((double)bk->entry_hhwater*100)/bk->entry_all,
            bk->entry_hwater,   ((double)bk->entry_hwater*100)/bk->entry_all,
            bk->entry_lwater,   ((double)bk->entry_lwater*100)/bk->entry_all
        );
#endif

    sum += tmp_sum;
    mlist_add_elem("bucket bodies", (void*)NULL, tmp_sum);

    Trace(" require_total_memory_size %lld\n",
        (long long)require_total_memory_size); 


    /*
     * frags
     */
    {
        int esize;
        long long   mlist_get_total();

        esize = sizeof(skcache_frag_t)+xffl_size;
        frag_size = xffl_size;
        Trace("  esize %d (%d + %d)\n",
            esize, sizeof(skcache_frag_t),xffl_size);
        Trace("  memory total %lld %.2fM, used %lld %.2fM, rest %lld %.2fM\n",
            (long long)require_total_memory_size,
                fdM((long long)require_total_memory_size),
            mlist_get_total(),
                fdM(mlist_get_total()),
            (long long)require_total_memory_size - mlist_get_total(),
                fdM((long long)require_total_memory_size - mlist_get_total()) );
        Trace("  xffl_num before %d ; %lld\n",
            xffl_num, (long long)xffl_num*esize);
        xffl_num =
            ((long long) require_total_memory_size - mlist_get_total() )
                / esize;
        Trace("  xffl_num after  %d ; %lld\n",
            xffl_num, (long long)xffl_num*esize);

        got = alloc_frag_freelist(&z_frag_freelist, 
                xffl_num, xffl_size);
        sum += got;
#if 0
        mlist_add_elem("frags", (void*)z_frag_freelist, got);
#endif
    }
    if(got<0) {
        Error("Sorry, no memory for frag free list");
        return -2;
    }

    frag_all = frag_free = xffl_num;
    frag_use  = 0;

    frag_hhwater = (int)(require_frag_hhwater*frag_all/100);
    frag_hwater  = (int)(require_frag_hwater*frag_all/100);
    frag_lwater  = (int)(require_frag_lwater*frag_all/100);


#if 1
    Trace(" frags:\n\
  the number of frags   %d\n\
  watermark             hhigh %d%% %d, high %d%% %d, low %d%% %d\n",
        frag_free,
        require_frag_hhwater, frag_hhwater,
        require_frag_hwater,  frag_hwater,
        require_frag_lwater,  frag_lwater);
#if 1
    Trace("\
  Size of frag cell is %d+%d = %d\n",
        sizeof(skcache_frag_t), xffl_size, sizeof(skcache_frag_t)+xffl_size);
#endif
#endif


    skcache_keeptime_token_init = require_keeptime_token_val;
    if(skcache_keeptime_token_init<2) {
        skcache_keeptime_token_init = 2;
    }
    skcache_keeptime_token = skcache_keeptime_token_init;

    Trace("  keeptime updating token    %d\n", skcache_keeptime_token_init);
    Trace("  keeptime updating interval %d\n", skcache_keeptime_token_init*2);

    Trace("\n\
Magic:\n\
  decimal   %3d %3d\n\
  hex       %3x %3x\n\
  charactor '%c%c'\n", 
    MAGIC_0, MAGIC_1, MAGIC_0, MAGIC_1, MAGIC_0, MAGIC_1);



#if 0
    disk_sweep_interval = require_disk_sweep_interval;
#endif

#if 0
    Trace("skcache_init: allocate %lld byte(s) %dK %.2fM %.2fG\n",
        (long long) sum, (long)dK(sum), fdM(sum), fdG(sum));
#endif

    sim_stat_U_alloc_new(&msweep_time,      "memory sweep time [msec]",
        1, K(10), msweep_time_count);
    sim_stat_U_alloc_new(&msweep_time_short, "memory sweep time (short) [msec]",
        1, K(10), msweep_time_short_count);

    sim_stat_U_alloc_new(&compaction_time,  "compaction time [msec]",
        100, K(2),   compaction_time_count);
    sim_stat_U_alloc_new(&checkmem_time,    "checkmem time [msec]",
        100, K(2),   checkmem_time_count);
    sim_hist_alloc_new(&fratio_hist,        "frag ratio history",
        MAX(skcache_keeptime_ave_range ,3),  fratio_hist_area);

#if 1
    {
        bk = &skcache_bucket_base[0];
    Trace("\n\
\n\
cache factors:\n\
    %-12s %8s %8s %8s %8s\n\
    %-12s %8d %8d %8d %8d\n\
    %-12s %8d %8d %8d %8d\n\
    %-12s %8d %8d %8d %8d\n\
\n",
    "name", "all", "hhwater", "hwater", "lwater",
    "frag", frag_all, frag_hhwater, frag_hwater, frag_lwater,
    "entry", bk->entry_all*radix_h1, bk->entry_hhwater*radix_h1, bk->entry_hwater*radix_h1, bk->entry_lwater*radix_h1,
    "  @bucket", bk->entry_all, bk->entry_hhwater, bk->entry_hwater, bk->entry_lwater);
    }
    Trace("  reaper will remove %d per sweep\n",
        bk->entry_hwater+1 - bk->entry_lwater);
    Trace("\n\n");
#endif

    return sum;
}

void
skcache_mem_init()
{
    TRACE_RECEIPT;
}
